#include <string>
#include <set>
#include <map>
#include <utility>

#define STRICT
#include <windows.h>
#include <tlhelp32.h>

#include "dbghelp.hpp"

#include "iat.hpp"
#include "util.hpp"

using namespace std;

namespace{
    typedef set<string> ExcModSet;
    typedef map<string, FARPROC> HookMap;
    typedef map<string, HookMap> PatchMap;

    template<typename Ptr, typename Base, typename Off>
    Ptr make_ptr(Base base, Off off)
    {
        return reinterpret_cast<Ptr>(reinterpret_cast<DWORD>(base)+reinterpret_cast<DWORD>(off));
    }

    bool patch_mod(const PatchMap& patch_map, HMODULE hMod)
    {
        ulong imp_size;
        PIMAGE_IMPORT_DESCRIPTOR imp = static_cast<PIMAGE_IMPORT_DESCRIPTOR>(
            dbghelp::ImageDirectoryEntryToData(hMod, TRUE, IMAGE_DIRECTORY_ENTRY_IMPORT, &imp_size));
        /**
         * ntdll.dll など一部のモジュールは IMAGE_IMPORT_DESCRIPTOR を持たな
         * いので、その場合は何もせず true を返す
         */
        if(!imp) return true;

        /**
         * モジュールがインポートしているDLLを全て列挙し、フック対象ならば
         * IATパッチを行う
         */
        for( ; imp->Name; ++imp){
            // フック対象DLLでなければ次へ
            string mod_name = make_ptr<char*>(hMod, imp->Name);
            PatchMap::const_iterator patch = patch_map.find(util::str_lower_ascii(mod_name));
            if(patch == patch_map.end()) continue;

            // フック対象DLLの場合、HookMap の全関数についてIATパッチを行う
            const HookMap& hook_map = patch->second;
            PIMAGE_THUNK_DATA iat = make_ptr<PIMAGE_THUNK_DATA>(hMod, imp->FirstThunk);
            PIMAGE_THUNK_DATA intb = make_ptr<PIMAGE_THUNK_DATA>(hMod, imp->OriginalFirstThunk);
            for( ; iat->u1.Function; ++iat, ++intb){
                // 序数でインポートされているものは名前がないので無視
                if(!IMAGE_SNAP_BY_ORDINAL(intb->u1.Ordinal)){
                    // フック対象関数でなければ次へ
                    PIMAGE_IMPORT_BY_NAME impbn = make_ptr<PIMAGE_IMPORT_BY_NAME>(hMod, intb->u1.AddressOfData);
                    string name = reinterpret_cast<char*>(impbn->Name);
                    HookMap::const_iterator hook = hook_map.find(name);
                    if(hook == hook_map.end()) continue;

                    /**
                     * IATパッチ処理
                     *
                     * IATは書き込み禁止になっている可能性があるので、プロテ
                     * クトを変更する(おそらく、DLLに含まれる関数が一度でも
                     * 使われていると書き込み禁止になるものと思われる)
                     */
                    FARPROC func = hook->second;
                    DWORD protect_orig;
                    if(!VirtualProtect(&iat->u1.Function, sizeof(iat->u1.Function), PAGE_READWRITE, &protect_orig))
                        return false;
                    iat->u1.Function = reinterpret_cast<DWORD>(func);
                    if(!VirtualProtect(&iat->u1.Function, sizeof(iat->u1.Function), protect_orig, &protect_orig))
                        return false;
                }
            }
        }

        return true;
    }
}

namespace IAT{
    bool patch(const Patch* patches, const char** excmods)
    {
        HANDLE hSnap = CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, 0);
        if(hSnap == INVALID_HANDLE_VALUE) return false;

        ExcModSet excmod_set;
        for(const char** p = excmods; *p; ++p)
            excmod_set.insert(util::str_lower_ascii(*p));
        PatchMap patch_map;
        for(const Patch* p = patches; p->dll; ++p){
            HookMap hook_map;
            for(const Hook* q = p->hooks; q->name; ++q)
                hook_map.insert(make_pair(q->name, q->func));
            // MinGW GCC 3.4.4 だと stl_tree.h 内でwarningが出るが、多分問題はないはず
            patch_map.insert(make_pair(util::str_lower_ascii(p->dll), hook_map));
        }

        /**
         * プロセス内の全てのモジュールについてIATパッチを行う
         * ただし excmod_set に含まれるモジュールは除外
         */
        MODULEENTRY32 me32 = { sizeof(MODULEENTRY32) };
        for(BOOL more = Module32First(hSnap, &me32); more; more = Module32Next(hSnap, &me32)){
            if(excmod_set.find(util::str_lower_ascii(me32.szModule)) != excmod_set.end()) continue;
            if(!patch_mod(patch_map, me32.hModule)) return false;
        }

        CloseHandle(hSnap);
        return true;
    }
}
