#include <fstream>
#include <string>
#include <vector>
#include <map>
#include <utility>
#include <stdint.h>

#include "tas.hpp"
#include "movie.hpp"
#include "util.hpp"
#include "types.hpp"

using namespace std;

namespace{
    bool chkread(streambuf* inb, char* buf, streamsize n)
    {
        return inb->sgetn(buf, n) == n;
    }

    bool chkread_u32(streambuf* inb, ulong& value)
    {
        char buf[4];
        if(!chkread(inb, buf, NELEMS(buf))) return false;
        value = util::unpack<uint32_t>(buf);
        return true;
    }

    bool chkread_double(streambuf* inb, double& value)
    {
        char buf[sizeof(double)];
        if(!chkread(inb, buf, NELEMS(buf))) return false;
        value = util::unpack<double>(buf);
        return true;
    }

    bool chkwrite(streambuf* outb, char* buf, streamsize n)
    {
        return outb->sputn(buf, n) == n;
    }

    bool chkwrite_u32(streambuf* outb, ulong value)
    {
        char buf[4];
        util::pack(buf, static_cast<uint32_t>(value));
        return chkwrite(outb, buf, NELEMS(buf));
    }

    bool chkwrite_double(streambuf* outb, double value)
    {
        char buf[sizeof(double)];
        util::pack(buf, value);
        return chkwrite(outb, buf, NELEMS(buf));
    }
}

Movie::Input::Input() : value_(0) {}
Movie::Input::Input(ulong value) : value_(value) {}
Movie::Input::Input(const char* packed) : value_(util::unpack<uint32_t>(packed)) {}
Movie::Input::Op Movie::Input::get(Button btn) const { return static_cast<Op>((value_>>(2*btn))&0x03); }
void Movie::Input::set(Button btn, Op op)
{
    value_ &= ~(0x03<<(2*btn));
    value_ |= op<<(2*btn);
}
ulong Movie::Input::value() const { return value_; }
void Movie::Input::pack(char* buf) const { util::pack(buf, value_); }

Movie::State::State(ulong rand, double x, double y, double vx, double vy)
    : rand_(rand), x_(x), y_(y), vx_(vx), vy_(vy) {}

bool Movie::State::operator==(const State& that) const
{
    return rand_==that.rand_ && x_==that.x_ && y_==that.y_ && vx_==that.vx_ && vy_==that.vy_;
}

bool Movie::State::operator!=(const State& that) const { return !operator==(that); }

bool Movie::load(Movie& movie, const string& path_rfr)
{
    string path_st = path_rfr + ".st";
    ifstream in_rfr(path_rfr.c_str(), ios_base::in | ios_base::binary);
    if(!in_rfr) return false;
    ifstream in_st(path_st.c_str(), ios_base::in | ios_base::binary);
    if(!in_st) return false;

    streambuf* inb_rfr = in_rfr.rdbuf();
    streambuf* inb_st = in_st.rdbuf();

#define CHKREAD(i,b,n) if(!chkread(i,b,n)) { return false; }
#define CHKREAD_U32(i,v) if(!chkread_u32(i,v)) { return false; }
#define CHKREAD_DOUBLE(i,v) if(!chkread_double(i,v)) { return false; }
    ulong stage, seed, frame_count;
    CHKREAD_U32(inb_rfr, stage);
    CHKREAD_U32(inb_rfr, seed);
    CHKREAD_U32(inb_rfr, frame_count);
    Movie tmp(stage, seed);
    for(ulong i = 0; i < frame_count; ++i){
        ulong value;
        CHKREAD_U32(inb_rfr, value);
        tmp.append(Input(value));
    }

    ulong rand;
    double x, y, vx, vy;
    CHKREAD_U32(inb_st, rand);
    CHKREAD_DOUBLE(inb_st, x);
    CHKREAD_DOUBLE(inb_st, y);
    CHKREAD_DOUBLE(inb_st, vx);
    CHKREAD_DOUBLE(inb_st, vy);
    tmp.setState(State(rand, x, y, vx, vy));
    ulong f;
    while(chkread_u32(inb_st, f)){
        char buf[4];
        CHKREAD(inb_st, buf, NELEMS(buf));
        int ex_wait = util::unpack<uchar>(buf+0);
        tmp.extraWaits_.insert(make_pair(f, ex_wait));
    }
#undef CHKREAD_DOUBLE
#undef CHKREAD_U32
#undef CHKREAD

    in_st.close();
    in_rfr.close();

    movie = tmp;
    return true;
}

bool Movie::save(const Movie& movie, const string& path_rfr)
{
    string path_st = path_rfr + ".st";
    ofstream out_rfr(path_rfr.c_str(), ios_base::out | ios_base::binary);
    if(!out_rfr) return false;
    ofstream out_st(path_st.c_str(), ios_base::out | ios_base::binary);
    if(!out_st) return false;

    streambuf* outb_rfr = out_rfr.rdbuf();
    streambuf* outb_st = out_st.rdbuf();

#define CHKWRITE(o,b,n) if(!chkwrite(o,b,n)) { return false; }
#define CHKWRITE_U32(o,v) if(!chkwrite_u32(o,v)) { return false; }
#define CHKWRITE_DOUBLE(o,v) if(!chkwrite_double(o,v)) { return false; }
    State state = movie.getState();
    CHKWRITE_U32(outb_st, state.rand_);
    CHKWRITE_DOUBLE(outb_st, state.x_);
    CHKWRITE_DOUBLE(outb_st, state.y_);
    CHKWRITE_DOUBLE(outb_st, state.vx_);
    CHKWRITE_DOUBLE(outb_st, state.vy_);

    CHKWRITE_U32(outb_rfr, movie.getStage());
    CHKWRITE_U32(outb_rfr, movie.getSeed());
    CHKWRITE_U32(outb_rfr, movie.getFrameCount());
    for(ulong i = 0; i < movie.getFrameCount(); ++i){
        ulong value = movie.get(i).value();
        CHKWRITE_U32(outb_rfr, value);
        int ex_wait = movie.getExtraWait(i);
        if(ex_wait){
            char buf[4] = {0};
            util::pack(buf+0, static_cast<uchar>(ex_wait));
            CHKWRITE_U32(outb_st, i);
            CHKWRITE(outb_st, buf, NELEMS(buf));
        }
    }
#undef CHKWRITE_DOUBLE
#undef CHKWRITE_U32
#undef CHKWRITE

    out_st.close();
    out_rfr.close();

    return true;
}

Movie::Movie(ulong stage, ulong seed)
    : stage_(stage), seed_(seed), inputs_(), extraWaits_(), state_(0, 0.0, 0.0, 0.0, 0.0) {}
Movie::Movie(const Movie& that)
    : stage_(that.stage_), seed_(that.seed_), inputs_(that.inputs_), extraWaits_(that.extraWaits_), state_(that.state_) {}

ulong Movie::getStage() const { return stage_; }
ulong Movie::getSeed() const { return seed_; }
ulong Movie::getFrameCount() const { return inputs_.size(); }

Movie::Input Movie::get(ulong frame) const { return inputs_[frame]; }
void Movie::append(const Input& input) { inputs_.push_back(input); }
void Movie::truncate(ulong frame) { inputs_.resize(frame); }

int Movie::getExtraWait(ulong frame) const
{
    return util::map_get(extraWaits_, frame, 0);
}

Movie::State Movie::getState() const { return state_; }
void Movie::setState(const State& state) { state_ = state; }

Movie::Input::Button& operator++(Movie::Input::Button& btn)
{
    return (btn = static_cast<Movie::Input::Button>(btn+1));
}
