#include <iostream>
#include <string>
/**
 * C++には stdint.h はないのだが、uint32_t を使いたいのでインクルードしてしまう
 * どうせMinGW専用なので問題ないだろう
 */
#include <stdint.h>

#define STRICT
#include <windows.h>
#include <mmsystem.h>

#include "dbghelp.hpp"
#include "iat.hpp"

#include "tas.hpp"
#include "game.hpp"
#include "movie.hpp"
#include "ui.hpp"
#include "throttle.hpp"
#include "util.hpp"
#include "types.hpp"

using namespace std;

namespace{
    /**
     * ステートロード中は描画は必要ないので、BitBlt() を無効化すれば速くなる
     * のでは、という発想。しかし実際にはほとんど効果がないようなので、現状
     * では bitblt_on は常に true となっている。
     */
    volatile bool bitblt_on = true;
    BOOL WINAPI my_BitBlt(HDC hdcDest, int nXDest, int nYDest, int nWidth, int nHeight,
                          HDC hdcSrc, int nXSrc, int nYSrc, DWORD dwRop)
    {
        if(bitblt_on)
            return BitBlt(hdcDest, nXDest, nYDest, nWidth, nHeight, hdcSrc, nXSrc, nYSrc, dwRop);
        else
            return TRUE;
    }

#define APICAST(f) reinterpret_cast<FARPROC>(f)
    const IAT::Hook IAT_HOOKS_WINMM[] = {
        { "timeGetTime", APICAST(Throttle::my_timeGetTime) },
        { 0, 0 }
    };
    const IAT::Hook IAT_HOOKS_GDI32[] = {
        { "BitBlt", APICAST(my_BitBlt) },
        { 0, 0 }
    };
#undef APICAST
    const IAT::Patch IAT_PATCHES[] = {
        { "winmm.dll", IAT_HOOKS_WINMM },
        { "gdi32.dll", IAT_HOOKS_GDI32 },
        { 0, 0 }
    };
    const char* IAT_EXCMOD[] = {
        "tasryoko.dll",
        0
    };

    const uint BTN2KEY[] = {
        VK_LEFT, VK_RIGHT, VK_UP, VK_DOWN, 'Z', 'X', 'C'
    };

    HINSTANCE hInstDLL_ = 0;

    bool paused_ = false;
    ulong frameWait_ = TAS::FRAME_WAIT_DEF;

    Movie::Input input_;
    bool btnState_[Movie::Input::BTN_LAST+1] = { false };

    ulong frameCount_ = 0;
    TAS::MovieMode movieMode_ = TAS::MOVIE_NONE;
    Movie movie_(0, 0);
    string moviePath_("");

    /**
     * 乱数やステージのアドレスは基本的に変化しない(プログラムがロードされる
     * アドレスは稀に変化するらしいので、100%ではないと思われる)。DLLインジェ
     * クション方式なので、DLLのサイズが変わっても不変のはず。
     */
    ulong addrRand_ = 0x00B31EA4;
    ulong addrStage_ = 0x00BFB8D8;
    ulong addrPosX_ = 0;

    ulong readMemS32(ulong addr)
    {
        char buf[4];
        Game::readMem(addr, buf, NELEMS(buf));
        return util::unpack<int32_t>(buf);
    }

    ulong readMemU32(ulong addr)
    {
        char buf[4];
        Game::readMem(addr, buf, NELEMS(buf));
        return util::unpack<uint32_t>(buf);
    }

    double readMemDouble(ulong addr)
    {
        char buf[sizeof(double)];
        Game::readMem(addr, buf, NELEMS(buf));
        return util::unpack<double>(buf);
    }

    void writeMemU32(ulong addr, ulong value)
    {
        char buf[4];
        util::pack(buf, value);
        Game::writeMem(addr, buf, NELEMS(buf));
    }

    Movie::Input doInput(Movie::Input input)
    {
        /**
         * 押し始めフラグをきちんと立ててやらないと、ゲーム側のリプレイ再生
         * でムービーが正しく認識されないので、適当に補正する
         */
        Movie::Input input_mod = input;
        for(Movie::Input::Button btn = Movie::Input::BTN_FIRST; btn <= Movie::Input::BTN_LAST; ++btn){
            Movie::Input::Op op = input.get(btn);
            if(btnState_[btn] && op == Movie::Input::PRESS ||
               !btnState_[btn] && op == Movie::Input::RELEASE){
                continue;
            }
            else if(!btnState_[btn]){
                // RELEASEPRESS は PRESS と同じ扱いにする
                input_mod.set(btn, Movie::Input::RELEASEPRESS); // 押し始めフラグ補正
                Game::sendKeyPress(BTN2KEY[btn]);
                btnState_[btn] = true;
            }
            else if(btnState_[btn] && op == Movie::Input::RELEASE){
                Game::sendKeyRelease(BTN2KEY[btn]);
                btnState_[btn] = false;
            }
            else{ // btnState_[btn] && op == Input::RELEASEPRESS
                Game::sendKeyRelease(BTN2KEY[btn]);
                Game::sendKeyPress(BTN2KEY[btn]);
            }
        }
        return input_mod;
    }

    /**
     * ゲーム側との同期をとるためのウェイト処理
     * (入力が反映されないうちにフレームを進めてしまうとdesyncするので)
     */
    void advSync()
    {
        /**
         * 以前考えた方法(自機座標などの変化を監視)
         * これでは結構な確率でdesyncが起こる。
         */
#if 0
        /**
         * Movie::State が変化したら入力が反映されたとみなす
         * ただし全く変化しないフレームもありうるので、50msのタイムアウトを
         * 設ける(タイムアウトが少し長いかもしれないが、TASならば状態が全く
         * 変化しないフレームは稀なのでさほど問題ではないだろう)
         *
         * Movie::State が変化したといっても、まだ演算途中であることも考えら
         * れるので、さらに16msのウェイトを入れる。これでdesyncはほぼ防げる
         * ようだ(8msだとたまにdesyncする)
         */
        Movie::State st_prev = TAS::genMovieState();
        ulong time_prev = timeGetTime();
        Throttle::frameadv();
        while(TAS::genMovieState() == st_prev && timeGetTime()-time_prev < 50) Throttle::waitSleep(0);
        Throttle::waitSleep(16);
#endif
        /**
         * 新手法(timeGetTime() が1回呼び出されるまで待つ)
         *
         * APIトレースを見る限り、ゲーム側のメインループは以下のような形になっ
         * ていると想像される:
         *   for(;;){
         *     ... // 区間A
         *     timeGetTime();
         *     ... // 区間B
         *     timeGetTime();
         *     ... // 区間C
         *   }
         * ということは、timeGetTime() が3回呼び出されるのを待てばループが1
         * 回りするので、ボタン入力は確実に反映されると考えられる。
         *
         * と考えていたのだが、どうも実際には1回呼ばれるまで待てば十分らしい
         * (これでもdesyncしないっぽい)。理由はよくわからないが、とにかく1回
         * 待てば入力は確実に反映されると見ていいようだ。
         *
         * これでもごく稀にdesyncするが、実用上問題ないレベルだと思う。
         */
        //Throttle::waitTGT(3);
        Throttle::waitTGT(1);
        Throttle::frameadv();
        /**
         * 毎フレーム上記のウェイト処理を入れるのであれば、
         * Throttle::frameadv() 後に Throttle::waitTGT() を呼ぶ必要はない。
         * ただし、その場合UI側に自機座標などが反映されるのが1F遅れることが
         * あるので、ムービー再生中などは advSync() 後に waitTGT() を呼び出
         * す方がいいだろう。逆に、ステートロード時は最終フレームだけ
         * waitTGT() を呼び出してフレーム処理を完了させれば十分なはず。
         *
         * なお、自機座標などをUIに反映させるためには waitTGT(3) が必要なよ
         * うだ(waitTGT(2) だとたまに反映されない模様)。これに関しては上で述
         * べた仮説が当たっているのかもしれない。
         */
        //Throttle::waitTGT(3);
    }

    bool init(HINSTANCE hInstDLL)
    {
        hInstDLL_ = hInstDLL;

        if(!dbghelp::init()) return false;
        if(!IAT::patch(IAT_PATCHES, IAT_EXCMOD)) return false;

        /**
         * このスレッドの優先度を下げる
         * desyncの軽減を期待してのものだが、どれほど効果があるかは未検証
         */
        if(!SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_LOWEST)) return false;

        if(!Game::init()) return false;
        if(!UI::init("Press Ctrl-P to setup addresses")) return false;

        return true;
    }

    bool fin()
    {
        bool ret = true;
        if(!UI::fin()) ret = false;
        if(!Game::fin()) ret = false;
        if(!dbghelp::fin()) ret = false;
        return ret;
    }
}

namespace TAS{
    HINSTANCE getHInstDLL() { return hInstDLL_; }

    void frameadv()
    {
        if(getMovieMode() == MOVIE_PLAY && getFrameCount() >= getMovie().getFrameCount())
            setMovieMode(MOVIE_NONE);

        Movie::Input input_mod = doInput(getInput());
        if(getMovieMode() != MOVIE_NONE){
            /**
             * 途中でBGMが変わるステージの場合、BGMの変わり目で余分にウェイ
             * トを入れてやらないとdesyncするので、ムービー側で追加ウェイト
             * を記録して再生時に反映させる。
             * BGMの変わり目では waitTGT(24) が必要なようだ
             */
            int ex_wait = getMovie().getExtraWait(getFrameCount());
            if(ex_wait)
                Throttle::waitTGT(ex_wait);
            advSync();
            /**
             * desyncを防ぐだけならこれは不要なはずだが、ある程度待たないと
             * 自機座標などが更新されない。ムービー再生中は正しい座標を表示
             * して欲しいので、一応待つことにする。
             * waitTGT(2) では足りない模様
             */
            Throttle::waitTGT(3);
        }
        else{
            Throttle::frameadv();
            Throttle::waitSleep(getFrameWait());
        }

        if(getMovieMode() == MOVIE_REC)
            movie_.append(input_mod); // 補正した入力をムービーに記録

        ++frameCount_;
        if(getMovieMode() == MOVIE_PLAY && getFrameCount() >= getMovie().getFrameCount()){
            setPause(true);
            if(genMovieState() != getMovie().getState())
                MessageBox(0, "DESYNC!!!", "tasryoko", MB_OK | MB_ICONERROR);
        }
        UI::invalidate();
    }

    bool isPaused() { return paused_; }
    void setPause(bool paused)
    {
        paused_ = paused;
        UI::invalidate();
    }
    ulong getFrameWait() { return frameWait_; }
    void setFrameWait(ulong ms) { frameWait_ = ms; }

    void retryStage()
    {
        const uint COORD_RETRY[2] = { 550, 260 };
        const uint COORD_OK[2] = { 270, 255 };

        bool paused_orig = isPaused();
        setPause(true);

        /**
         * 操作間に逐一ウェイトを入れてやらないとフレームがずれることがある。
         * 毎回 waitTGT(3) すれば大丈夫っぽい
         */

        // 1回画面をクリック。落下直後のフェードアウトなどはこれでキャンセルできると思う
        Throttle::waitTGT(3);
        Game::sendMouseClickL(COORD_OK[0], COORD_OK[1]);

        // 「フィールドやり直し」->「はい」
        Throttle::waitTGT(3);
        Game::sendMouseClickR(COORD_RETRY[0], COORD_RETRY[1]);
        Throttle::waitTGT(3);
        Game::sendMouseClickL(COORD_RETRY[0], COORD_RETRY[1]);
        Throttle::waitTGT(3);
        Game::sendMouseClickL(COORD_OK[0], COORD_OK[1]);

#if 1
        // 1F進める
        Throttle::waitTGT(3);
        Throttle::frameadv();
        // 1回画面をクリック。これをやらないと1Fずれる模様
        Throttle::waitTGT(3);
        Game::sendMouseClickL(COORD_OK[0], COORD_OK[1]);
        // 1F進める
        Throttle::waitTGT(3);
        Throttle::frameadv();
        // 再度画面をクリックすると0F目となる
        // 多分これでずれないはず
        Throttle::waitTGT(3);
        Game::sendMouseClickL(COORD_OK[0], COORD_OK[1]);
#endif
        /**
         * 乱数値を参照してフレームずれを防ぐ方法も考えたが、これは完璧に機
         * 能するか怪しい上に、機能しなかった場合は延々とフレームが進んでし
         * まうので却下
         */
#if 0
        while(getRand() + Throttle::FRAME_TIME != Throttle::getUptime()){
            Throttle::waitSleep(50);
            Throttle::frameadv();
            Throttle::waitSleep(50);
            Game::sendMouseClickL(COORD_OK[0], COORD_OK[1]);
        }
#endif
        Throttle::waitTGT(3);

        frameCount_ = 0;
        input_ = Movie::Input(static_cast<ulong>(0));
        for(uint i = 0; i < NELEMS(btnState_); ++i) btnState_[i] = false;
        setPause(paused_orig);
        UI::invalidate();
    }

    Movie::Input getInput()
    {
        return getMovieMode()==MOVIE_PLAY ? getMovie().get(getFrameCount()) : input_;
    }
    Movie::Input::Op getInput(Movie::Input::Button btn) { return getInput().get(btn); }
    void setInput(Movie::Input input)
    {
        if(getMovieMode() == MOVIE_PLAY) return;
        input_ = input;
        UI::invalidate();
    }
    void setInput(Movie::Input::Button btn, Movie::Input::Op op)
    {
        if(getMovieMode() == MOVIE_PLAY) return;
        input_.set(btn, op);
        UI::invalidate();
    }

    MovieMode getMovieMode() { return movieMode_; }
    void setMovieMode(MovieMode mode)
    {
        movieMode_ = mode;
        UI::invalidate();
    }
    const Movie& getMovie() { return movie_; }
    void setMovie(const Movie& movie) { movie_ = movie; }
    const string& getMoviePath() { return moviePath_; }
    void setMoviePath(const string& path) { moviePath_ = path; }

    Movie::State genMovieState()
    {
        double x, y, vx, vy;
        getPos(x, y);
        getVel(vx, vy);
        return Movie::State(getRand(), x, y, vx, vy);
    }

    bool stateLoad(const Movie& state)
    {
        if(state.getStage() != getStage()) return false;
        if(getMovieMode() != MOVIE_NONE && state.getSeed() != getMovie().getSeed()) return false;
        if(getMovieMode() == MOVIE_PLAY && state.getFrameCount() > getMovie().getFrameCount()) return false;

#if 0
        if(getMovieMode() == MOVIE_REC) movie_.truncate(0);
#endif
        retryStage();
        setRand(state.getSeed());

        /**
         * 描画を無効化すれば速くなるのでは、と思ったが、ほとんど効果がない
         * ようなのでやめた。
         */
        //bitblt_on = false;
        Movie::Input input;
        for(ulong i = 0; i < state.getFrameCount(); ++i){
            input = state.get(i);
            Movie::Input input_mod = doInput(input);
            int ex_wait = state.getExtraWait(i);
            if(ex_wait)
                Throttle::waitTGT(ex_wait);
            advSync();
#if 0
            if(getMovieMode() == MOVIE_REC)
                movie_.append(input_mod);
#endif
            ++frameCount_;
        }
        bitblt_on = true;
        /**
         * 最終フレームの処理が行われるまで待ち、自機座標などを更新させる
         * waitTGT(2) では足りない模様
         */
        Throttle::waitTGT(3);

        if(genMovieState() != state.getState())
            MessageBox(0, "DESYNC!!!", "tasryoko", MB_OK | MB_ICONERROR);
        if(getMovieMode() != MOVIE_PLAY) setInput(input);
        if(getMovieMode() == MOVIE_REC) movie_ = state;
        UI::invalidate();
        return true;
    }

    bool stateSave(const string& path)
    {
        // ムービーなしの場合、そもそも入力を記録してないのでセーブできない
        if(getMovieMode() == MOVIE_NONE) return false;

        Movie movie = getMovie();
        movie.truncate(getFrameCount());
        movie.setState(genMovieState());
        return Movie::save(movie, path);
    }

    void getPos(double& x, double& y)
    {
        x = readMemDouble(getAddrPosX()+0x00);
        y = readMemDouble(getAddrPosX()+0x08);
    }
    void getVel(double& vx, double& vy)
    {
        vx = readMemDouble(getAddrPosX()+0x10);
        vy = readMemDouble(getAddrPosX()+0x18);
    }

    ulong getRand() { return readMemU32(addrRand_); }
    void setRand(ulong r)
    {
        writeMemU32(addrRand_, r);
        UI::invalidate();
    }
    ulong getStage() { return readMemU32(addrStage_); }
    ulong getFrameCount()
    {
        return frameCount_;
    }

    ulong getAddrRand() { return addrRand_; }
    void setAddrRand(ulong addr) { addrRand_ = addr; }

    ulong getAddrStage() { return addrStage_; }
    void setAddrStage(ulong addr) { addrStage_ = addr; }

    ulong getAddrPosX() { return addrPosX_; }
    void setAddrPosX(ulong addr) { addrPosX_ = addr; }

    bool searchAddrPosX(ulong& addr)
    {
        const double POSX_INI = 2880.0;
        const ulong ADDR_BEG = 0;
        const ulong ADDR_END = 0x05000000;
        const ulong PAGE_SIZE = 0x10000;
        const ulong ALIGN = 8;

        int count = 0;
        for(ulong base = ADDR_BEG; base < ADDR_END; base += PAGE_SIZE){
            char buf[PAGE_SIZE];
            if(!Game::readMem(base, buf, NELEMS(buf))) continue;
            for(ulong i = 0; i < PAGE_SIZE; i += ALIGN){
                double x = util::unpack<double>(buf+i);
                if(x == POSX_INI){
                    ++count;
                    addr = base+i;
                }
            }
        }

        return count == 1;
    }

    /**
     * 自機データアドレス未取得、または式神が存在しない場合、0が返る
     */
    ulong getAddrShiki()
    {
        if(!getAddrPosX()) return 0;
        ulong ptr = readMemU32(getAddrPosX() + 0xA8);
        return readMemU32(ptr);
    }

    void getShiki(Shiki& shiki)
    {
        ulong addr = getAddrShiki();
        if(!addr) return;

        shiki.x = readMemDouble(addr+0x30);
        shiki.y = readMemDouble(addr+0x38);
        shiki.vx = readMemDouble(addr+0x40);
        shiki.vy = readMemDouble(addr+0x48);
        shiki.mode = readMemU32(addr+0x68);
        shiki.mode_timer = readMemU32(addr+0x6C);
        shiki.bound_frame = readMemS32(addr+0x74);
        shiki.ob_count = readMemU32(addr+0x78);
    }

    uint tasMain(HINSTANCE hInstDLL)
    {
        if(!init(hInstDLL)){
            MessageBox(0, "Initialization FAILED", "tasryoko", MB_OK);
            return 1;
        }

        if(!Game::setHook()) MessageBox(0, "setHook() failed", "t", MB_OK);

        uint ret = UI::loop();
        if(ret != 0) MessageBox(0, "UI::loop() ABORTED", "tasryoko", MB_OK);

        if(!fin()) return 1;

        return ret;
    }
}
