#include <sstream>
#include <string>
#include <cstring>

#define WINVER 0x0500
#define _WIN32_WINNT 0x0500
#define STRICT
#include <windows.h>

#include "tas.hpp"
#include "movie.hpp"
#include "throttle.hpp"
#include "ui.hpp"
#include "hotkey.hpp"
#include "util.hpp"
#include "types.hpp"

using namespace std;

namespace{
    const COLORREF COLOR_RED = RGB(255, 0, 0);
    const COLORREF COLOR_BLACK = RGB(0, 0, 0);
    const COLORREF COLOR_WHITE = RGB(255, 255, 255);

    const char CLASS_NAME[] = "KSK_TASRYOKO_UI";
    const int WND_SIZE[2] = { 480, 360 };
    const char FONT_FACE[] = "FixedSys";
    const int FONT_SIZE = 12;
    const COLORREF BG_COLOR = COLOR_WHITE;
    const COLORREF TEXT_COLOR_DEF = COLOR_BLACK;
    const char* const INPUT_CH[] = { "<", ">", "^", "v", "1", "2", "3" };

    const int PREC_POS = 4;
    const int PREC_VEL = 4;

    HWND hWnd_ = 0;
    HFONT hFont_ = 0;
    string statusLine_("");

    void drawStr(HDC hDC, int x, int y, const string& s, COLORREF color = TEXT_COLOR_DEF)
    {
        SetTextColor(hDC, color);
        TextOut(hDC, x, y, s.c_str(), s.size());
    }

    void fillRect(HDC hDC, int x, int y, int w, int h, COLORREF color)
    {
        SetDCPenColor(hDC, color);
        SetDCBrushColor(hDC, color);
        Rectangle(hDC, x, y, x+w, y+h);
    }

    void fillEllipse(HDC hDC, int x, int y, int w, int h, COLORREF color)
    {
        SetDCPenColor(hDC, color);
        SetDCBrushColor(hDC, color);
        Ellipse(hDC, x, y, x+w, y+h);
    }

    void fillPoly(HDC hDC, const POINT* vert, int n, COLORREF color)
    {
        SetDCPenColor(hDC, color);
        SetDCBrushColor(hDC, color);
        Polygon(hDC, vert, n);
    }

    void paintUptime(HDC hDC)
    {
        drawStr(hDC, 0, 0, "uptime: " + util::to_dec(Throttle::getUptime()));
    }

    void paintRand(HDC hDC)
    {
        drawStr(hDC, 0, 18, "rand: " + util::to_dec(TAS::getRand()));
    }

    void paintStage(HDC hDC)
    {
        drawStr(hDC, 0, 36, "stage: " + util::to_dec(TAS::getStage()));
    }

    void paintPlayerInfo(HDC hDC)
    {
        double x, y, vx, vy;
        TAS::getPos(x, y);
        TAS::getVel(vx, vy);
        drawStr(hDC, 0, 72, "pos: (" + util::to_dec(x,PREC_POS) + ", " + util::to_dec(y,PREC_POS) + ")");
        drawStr(hDC, 0, 90, "vel: (" + util::to_dec(vx,PREC_VEL) + ", " + util::to_dec(vy,PREC_VEL) + ")");
    }

    void paintShikiInfo(HDC hDC)
    {
        TAS::Shiki shiki;
        TAS::getShiki(shiki);

        ulong timer_limit = 0;
        if(shiki.mode == 2)
            timer_limit = 16;
        else if(shiki.mode != 3 && shiki.bound_frame > 0)
            timer_limit = 48+shiki.bound_frame;

        drawStr(hDC, 0, 126, "shiki pos: (" +
                util::to_dec(shiki.x,PREC_POS) + ", " + util::to_dec(shiki.y,PREC_POS) + ")");
        drawStr(hDC, 0, 144, "shiki vel: (" +
                util::to_dec(shiki.vx,PREC_VEL) + ", " + util::to_dec(shiki.vy,PREC_VEL) + ")");
        drawStr(hDC, 0, 162, "shiki mode: " +
                util::to_dec(shiki.mode) + ", timer=" + util::to_dec(shiki.mode_timer) +
                (timer_limit ? "/"+util::to_dec(timer_limit) : "") + ", " +
                "ob=" + util::to_dec(shiki.ob_count) + "/30");
    }

    void paintFrameCount(HDC hDC)
    {
        string s = util::to_dec(TAS::getFrameCount());
        if(TAS::getMovieMode() == TAS::MOVIE_PLAY)
            s += " / " + util::to_dec(TAS::getMovie().getFrameCount());
        drawStr(hDC, 0, 282, s);
    }

    void paintPauseMark(HDC hDC)
    {
        fillRect(hDC, 404, 300, 8, 32, COLOR_RED);
        fillRect(hDC, 420, 300, 8, 32, COLOR_RED);
    }

    void paintPlayMark(HDC hDC)
    {
        const POINT poly_play[] = { {436,302}, {436,330}, {462,316} };
        fillPoly(hDC, poly_play, NELEMS(poly_play), COLOR_RED);
    }

    void paintRecMark(HDC hDC)
    {
        fillEllipse(hDC, 434, 302, 28, 28, COLOR_RED);
    }

    void paintInput(HDC hDC)
    {
        const Movie::Input& input = TAS::getInput();
        for(Movie::Input::Button btn = Movie::Input::BTN_FIRST; btn <= Movie::Input::BTN_LAST; ++btn){
            Movie::Input::Op op = input.get(btn);
            if(op != Movie::Input::RELEASE)
                drawStr(hDC, 0+(8*btn), 300, INPUT_CH[btn],
                        (op==Movie::Input::RELEASEPRESS ? COLOR_RED : TEXT_COLOR_DEF));
        }
    }

    void paintStatusLine(HDC hDC)
    {
        drawStr(hDC, 0, 318, statusLine_);
    }

    void paint(HDC hDC)
    {
        paintUptime(hDC);
        paintRand(hDC);
        paintStage(hDC);
        if(TAS::getAddrPosX()){
            paintPlayerInfo(hDC);
        }
        if(TAS::getAddrShiki()){
            paintShikiInfo(hDC);
        }
        paintFrameCount(hDC);

        if(TAS::isPaused()) paintPauseMark(hDC);
        switch(TAS::getMovieMode()){
        case TAS::MOVIE_NONE: break;
        case TAS::MOVIE_PLAY: paintPlayMark(hDC); break;
        case TAS::MOVIE_REC: paintRecMark(hDC); break;
        }

        paintInput(hDC);

        paintStatusLine(hDC);
    }

    bool wndHandler_WM_PAINT(HWND hWnd, WPARAM wp, LPARAM lp, LRESULT& ret)
    {
        PAINTSTRUCT ps;
        HDC hDC = BeginPaint(hWnd, &ps);
        SelectObject(hDC, GetStockObject(DC_PEN));
        SelectObject(hDC, GetStockObject(DC_BRUSH));
        HFONT hFontOld = static_cast<HFONT>(SelectObject(hDC, hFont_));
        paint(hDC);
        SelectObject(hDC, hFontOld);
        EndPaint(hWnd, &ps);
        ret = 0;
        return true;
    }

    bool wndHandler_WM_KEYDOWN(HWND hWnd, WPARAM wp, LPARAM lp, LRESULT& ret)
    {
        ret = 0;
        return Hotkey::press(hWnd, static_cast<uint>(wp));
    }

    bool wndHandler_WM_KEYUP(HWND hWnd, WPARAM wp, LPARAM lp, LRESULT& ret)
    {
        ret = 0;
        return Hotkey::release(hWnd, static_cast<uint>(wp));
    }

    bool wndHandler_WM_SYSKEYDOWN(HWND hWnd, WPARAM wp, LPARAM lp, LRESULT& ret)
    {
        ret = 0;
        return Hotkey::press(hWnd, static_cast<uint>(wp));
    }

    bool wndHandler_WM_SYSKEYUP(HWND hWnd, WPARAM wp, LPARAM lp, LRESULT& ret)
    {
        ret = 0;
        return Hotkey::release(hWnd, static_cast<uint>(wp));
    }

    long pt2lfHeight(long pt, HDC hDC)
    {
        return -MulDiv(pt, GetDeviceCaps(hDC, LOGPIXELSY), 72);
    }

    bool initFont(HWND hWnd)
    {
        LOGFONT lf;
        BYTE charset;

        if(!GetObject(GetStockObject(DEFAULT_GUI_FONT), sizeof(lf), &lf)) return false;
        charset = lf.lfCharSet;
        memset(&lf, 0, sizeof(lf));
        lf.lfCharSet = charset;
        strncpy(lf.lfFaceName, FONT_FACE, LF_FACESIZE-1);
        lf.lfFaceName[LF_FACESIZE-1] = '\0';
        HDC hDC = GetDC(hWnd);
        if(!hDC) return false;
        lf.lfHeight = pt2lfHeight(FONT_SIZE, hDC);
        ReleaseDC(hWnd, hDC);

        if(!(hFont_ = CreateFontIndirect(&lf))) return false;

        return true;
    }

    bool finFont()
    {
        return DeleteObject(hFont_);
    }

    bool wndHandler_WM_CREATE(HWND hWnd, WPARAM wp, LPARAM lp, LRESULT& ret)
    {
        if(!initFont(hWnd)) goto FAIL;

        ret = 0;
        return true;
FAIL:
        ret = -1;
        return true;
    }

    bool wndHandler_WM_DESTROY(HWND hWnd, WPARAM wp, LPARAM lp, LRESULT& ret)
    {
        finFont();

        ret = 0;
        return true;
    }

    LRESULT CALLBACK wndProc(HWND hWnd, UINT msg, WPARAM wp, LPARAM lp)
    {
        LRESULT ret;
        switch(msg){
#define MY_HANDLE_MSG(m) case m: if(wndHandler_##m (hWnd, wp, lp, ret)) return ret; break
            MY_HANDLE_MSG(WM_CREATE);
            MY_HANDLE_MSG(WM_DESTROY);
            MY_HANDLE_MSG(WM_KEYDOWN);
            MY_HANDLE_MSG(WM_KEYUP);
            MY_HANDLE_MSG(WM_SYSKEYDOWN);
            MY_HANDLE_MSG(WM_SYSKEYUP);
            MY_HANDLE_MSG(WM_PAINT);
#undef MY_HANDLE_MSG
        }
        return DefWindowProc(hWnd, msg, wp, lp);
    }

    bool initWndClass()
    {
        WNDCLASSEX wc;
        wc.cbSize = sizeof(wc);
        wc.style = CS_HREDRAW | CS_VREDRAW;
        wc.lpfnWndProc = wndProc;
        wc.cbClsExtra = 0;
        wc.cbWndExtra = 0;
        wc.hInstance = TAS::getHInstDLL();
        wc.hIcon = 0;
        wc.hIconSm = wc.hIcon;
        wc.hCursor = LoadCursor(0, IDC_ARROW);
        wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
        wc.lpszMenuName = 0;
        wc.lpszClassName = CLASS_NAME;
        return RegisterClassEx(&wc);
    }

    bool initWnd()
    {
        DWORD style = WS_OVERLAPPED;
        DWORD style_ex = 0;
        hWnd_ = CreateWindowEx(style_ex, CLASS_NAME, "tasryoko", style,
                              CW_USEDEFAULT, CW_USEDEFAULT, WND_SIZE[0], WND_SIZE[1],
                              0, 0, TAS::getHInstDLL(), 0);
        if(!hWnd_) return false;
        ShowWindow(hWnd_, SW_SHOWNORMAL);
        UpdateWindow(hWnd_);
        return true;
    }
}

namespace UI{
    bool init(const string& statusLine)
    {
        setStatusLine(statusLine);
        if(!initWndClass()) return false;
        if(!initWnd()) return false;
        return true;
    }

    bool fin() { return true; }

    uint loop()
    {
        for(;;){
            MSG msg;
            if(PeekMessage(&msg, 0, 0, 0, PM_NOREMOVE)){
                BOOL st = GetMessage(&msg, 0, 0, 0);
                if(st == -1) return 1; // ERROR
                if(st == 0) return msg.wParam; // WM_QUIT
                DispatchMessage(&msg);
            }
            else{
                if(!TAS::isPaused()) TAS::frameadv();
            }
        }
    }

    void invalidate()
    {
        InvalidateRect(hWnd_, 0, TRUE);
    }

    void update()
    {
        UpdateWindow(hWnd_);
    }

    void setStatusLine(const string& statusLine) { statusLine_ = statusLine; }
}
